@@ -1,6 +1,8 @@
 # coding: utf-8
+import cookielib
 from copy import copy
 from functools import partial
+from contextlib import contextmanager
 
 from werkzeug.local import LocalStack
 from flask import g, session, get_flashed_messages
@@ -127,6 +129,16 @@ class TestRequest(BaseTestRequest):
     ResponseClass = TestResponse
 
 
+class CookieJar(cookielib.CookieJar):
+    """CookieJar that always sets ASCII headers, even if cookies have
+    unicode parts such as name, value or path. It is necessary to make
+    :meth:`TestApp.session_transaction` work correctly.
+    """
+    def _cookie_attrs(self, cookies):
+        attrs = cookielib.CookieJar._cookie_attrs(self, cookies)
+        return map(str, attrs)
+
+
 class TestApp(BaseTestApp):
     """Extends :class:`webtest.TestApp` by adding few fields to responses:
 
@@ -147,24 +159,41 @@ class TestApp(BaseTestApp):
 
     .. attribute:: session
 
-        Dictionary containing session data.
+        Dictionary that contains session data.
 
     If exactly one template was used to render the response, it's name and context
     can be accessed using `response.template` and `response.context` properties.
 
+    If `app` config sets SERVER_NAME and HTTP_HOST is not specified in
+    `extra_environ`, :class:`TestApp` will also set HTTP_HOST to SERVER_NAME
+    for all requests to the app.
+
+    :param app: :class:`flask.Flask` instance
     :param db: :class:`flask.ext.sqlalchemy.SQLAlchemy` instance
     :param use_session_scopes: if specified, application performs each request
-                               within it's own separate session scopes
+                               within it's own separate session scope
     """
     RequestClass = TestRequest
 
-    def __init__(self, app, db=None, use_session_scopes=False, *args, **kwargs):
+    def __init__(self, app, db=None, use_session_scopes=False, cookiejar=None,
+                 extra_environ=None, *args, **kwargs):
         if use_session_scopes:
             assert db, ('`db` (instance of `flask.ext.sqlalchemy.SQLAlchemy`) '
                         'must be passed to use session scopes.')
         self.db = db
         self.use_session_scopes = use_session_scopes
-        super(TestApp, self).__init__(app, *args, **kwargs)
+
+        if extra_environ is None:
+            extra_environ = {}
+        if app.config['SERVER_NAME'] and 'HTTP_HOST' not in extra_environ:
+            extra_environ['HTTP_HOST'] = app.config['SERVER_NAME']
+
+        super(TestApp, self).__init__(app, extra_environ=extra_environ,
+                                      *args, **kwargs)
+        # cookielib.CookieJar defines __len__ and empty CookieJar evaluates
+        # to False in boolan context. That's why we explicitly compare
+        # `cookiejar` with None:
+        self.cookiejar = CookieJar() if cookiejar is None else cookiejar
 
     def do_request(self, *args, **kwargs):
         store = {}
@@ -194,3 +223,33 @@ class TestApp(BaseTestApp):
         response.flashes = store.get('flashes', [])
         response.contexts = dict(store.get('contexts', []))
         return response
+
+    @contextmanager
+    def session_transaction(self):
+        """When used in combination with a with statement this opens
+        a session transaction. This can be used to modify the session
+        that the test client uses. Once the with block is left the session
+        is stored back.
+
+        For example, if you use Flask-Login, you can log in a user using
+        this method::
+
+            with client.session_transaction() as sess:
+                sess['user_id'] = 1
+
+        Internally it uses :meth:`flask.testing.FlaskClient.session_transaction`.
+        """
+        with self.app.test_client() as client:
+            for cookie in self.cookiejar:
+                client.cookie_jar.set_cookie(cookie)
+
+            with client.session_transaction() as sess:
+                yield sess
+
+            for cookie in client.cookie_jar:
+                # Cookies from `client.cookie_jar` may contain unicode name
+                # and value. It would make WebTest linter (:mod:`webtest.lint`)
+                # throw assertion errors about unicode environmental
+                # variable (HTTP_COOKIE), but we use custom CookieJar that is
+                # aware of this oddity and always sets 8-bit headers.
+                self.cookiejar.set_cookie(cookie)
